(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('worker-timers'), require('midi-file-slicer'), require('@babel/runtime/helpers/defineProperty'), require('json-midi-message-encoder'), require('@babel/runtime/helpers/classCallCheck'), require('@babel/runtime/helpers/createClass'), require('rxjs')) :
    typeof define === 'function' && define.amd ? define(['exports', 'worker-timers', 'midi-file-slicer', '@babel/runtime/helpers/defineProperty', 'json-midi-message-encoder', '@babel/runtime/helpers/classCallCheck', '@babel/runtime/helpers/createClass', 'rxjs'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.midiPlayer = {}, global.workerTimers, global.midiFileSlicer, global._defineProperty, global.jsonMidiMessageEncoder, global._classCallCheck, global._createClass, global.rxjs));
})(this, (function (exports, workerTimers, midiFileSlicer, _defineProperty, jsonMidiMessageEncoder, _classCallCheck, _createClass, rxjs) { 'use strict';

    var createMidiFileSlicer = function createMidiFileSlicer(json) {
      return new midiFileSlicer.MidiFileSlicer({
        json: json
      });
    };

    var encodeMidiMessage = function encodeMidiMessage(event) {
      return new Uint8Array(jsonMidiMessageEncoder.encode(event));
    };

    var MidiPlayer = /*#__PURE__*/function () {
      function MidiPlayer(_ref) {
        var encodeMidiMessage = _ref.encodeMidiMessage,
          filterMidiMessage = _ref.filterMidiMessage,
          json = _ref.json,
          midiFileSlicer = _ref.midiFileSlicer,
          midiOutput = _ref.midiOutput,
          scheduler = _ref.scheduler;
        _classCallCheck(this, MidiPlayer);
        this._encodeMidiMessage = encodeMidiMessage;
        this._filterMidiMessage = filterMidiMessage;
        this._json = json;
        this._midiFileSlicer = midiFileSlicer;
        this._midiOutput = midiOutput;
        this._scheduler = scheduler;
        this._state = null;
      }
      return _createClass(MidiPlayer, [{
        key: "play",
        value: function play() {
          var _this = this;
          if (this._state !== null) {
            throw new Error('The player is currently playing.');
          }
          return new Promise(function (resolve, reject) {
            var schedulerSubscription = _this._scheduler.subscribe({
              error: function error(err) {
                return reject(err);
              },
              next: function next(_ref2) {
                var end = _ref2.end,
                  start = _ref2.start;
                if (_this._state === null) {
                  _this._state = {
                    endedTracks: 0,
                    offset: start,
                    resolve: resolve,
                    schedulerSubscription: null
                  };
                }
                _this._schedule(start, end, _this._state);
              }
            });
            if (_this._state === null) {
              schedulerSubscription.unsubscribe();
            } else {
              _this._state.schedulerSubscription = schedulerSubscription;
            }
          });
        }
      }, {
        key: "stop",
        value: function stop() {
          if (this._state === null) {
            throw new Error('The player is already stopped.');
          }
          this._stop(this._state);
        }
      }, {
        key: "_schedule",
        value: function _schedule(start, end, state) {
          var _this2 = this;
          var events = this._midiFileSlicer.slice(start - state.offset, end - state.offset);
          events.filter(function (_ref3) {
            var event = _ref3.event;
            return _this2._filterMidiMessage(event);
          }).forEach(function (_ref4) {
            var event = _ref4.event,
              time = _ref4.time;
            return _this2._midiOutput.send(_this2._encodeMidiMessage(event), start + time);
          });
          var endedTracks = events.filter(function (_ref5) {
            var event = _ref5.event;
            return MidiPlayer._isEndOfTrack(event);
          }).length;
          state.endedTracks += endedTracks;
          if (state.endedTracks === this._json.tracks.length) {
            this._stop(state);
          }
        }
      }, {
        key: "_stop",
        value: function _stop(state) {
          var resolve = state.resolve,
            schedulerSubscription = state.schedulerSubscription;
          schedulerSubscription === null || schedulerSubscription === void 0 ? void 0 : schedulerSubscription.unsubscribe();
          this._state = null;
          resolve();
        }
      }], [{
        key: "_isEndOfTrack",
        value: function _isEndOfTrack(event) {
          return 'endOfTrack' in event;
        }
      }]);
    }();

    function ownKeys(e, r) { var t = Object.keys(e); if (Object.getOwnPropertySymbols) { var o = Object.getOwnPropertySymbols(e); r && (o = o.filter(function (r) { return Object.getOwnPropertyDescriptor(e, r).enumerable; })), t.push.apply(t, o); } return t; }
    function _objectSpread(e) { for (var r = 1; r < arguments.length; r++) { var t = null != arguments[r] ? arguments[r] : {}; r % 2 ? ownKeys(Object(t), !0).forEach(function (r) { _defineProperty(e, r, t[r]); }) : Object.getOwnPropertyDescriptors ? Object.defineProperties(e, Object.getOwnPropertyDescriptors(t)) : ownKeys(Object(t)).forEach(function (r) { Object.defineProperty(e, r, Object.getOwnPropertyDescriptor(t, r)); }); } return e; }
    var createMidiPlayerFactory = function createMidiPlayerFactory(createMidiFileSlicer, scheduler) {
      return function (options) {
        var midiFileSlicer = createMidiFileSlicer(options.json);
        return new MidiPlayer(_objectSpread(_objectSpread({
          filterMidiMessage: function filterMidiMessage(event) {
            return 'controlChange' in event || 'noteOff' in event || 'noteOn' in event || 'programChange' in event;
          }
        }, options), {}, {
          encodeMidiMessage: encodeMidiMessage,
          midiFileSlicer: midiFileSlicer,
          scheduler: scheduler
        }));
      };
    };

    var INTERVAL = 500;
    var Scheduler = /*#__PURE__*/function () {
      function Scheduler(_clearInterval, _performance, _setInterval) {
        _classCallCheck(this, Scheduler);
        this._clearInterval = _clearInterval;
        this._performance = _performance;
        this._setInterval = _setInterval;
        this._intervalId = null;
        this._nextTick = 0;
        this._numberOfSubscribers = 0;
        this._subject = new rxjs.Subject();
      }
      return _createClass(Scheduler, [{
        key: "subscribe",
        value: function subscribe(observer) {
          var _this = this;
          this._numberOfSubscribers += 1;
          var currentTime = this._performance.now();
          if (this._numberOfSubscribers === 1) {
            this._start(currentTime);
          }
          // tslint:disable-next-line:deprecation
          var subscription = rxjs.merge(rxjs.of({
            end: this._nextTick + INTERVAL,
            start: currentTime
          }), this._subject).subscribe(observer);
          var unsubscribe = function unsubscribe() {
            _this._numberOfSubscribers -= 1;
            if (_this._numberOfSubscribers === 0) {
              _this._stop();
            }
            return subscription.unsubscribe();
          };
          return {
            unsubscribe: unsubscribe
          };
        }
      }, {
        key: "_start",
        value: function _start(currentTime) {
          var _this2 = this;
          this._nextTick = currentTime + INTERVAL;
          this._intervalId = this._setInterval(function () {
            if (_this2._performance.now() >= _this2._nextTick) {
              _this2._nextTick += INTERVAL;
              _this2._subject.next({
                end: _this2._nextTick + INTERVAL,
                start: _this2._nextTick
              });
            }
          }, INTERVAL / 10);
        }
      }, {
        key: "_stop",
        value: function _stop() {
          if (this._intervalId !== null) {
            this._clearInterval(this._intervalId);
          }
          this._intervalId = null;
        }
      }]);
    }();

    var scheduler = new Scheduler(workerTimers.clearInterval, performance, workerTimers.setInterval);
    var createMidiPlayer = createMidiPlayerFactory(createMidiFileSlicer, scheduler);
    var create = function create(options) {
      return createMidiPlayer(options);
    };

    exports.create = create;

}));
